import curses
import json
from abc import ABC
from collections.abc import Iterable
from dataclasses import dataclass
from decimal import Decimal
from math import log10
from typing import override

from tjex.config import config
from tjex.curses_helper import Region
from tjex.point import Point
from tjex.table import CellFormatter, Table

type Json = str | int | float | bool | list[Json] | dict[str, Json] | None


@dataclass(frozen=True)
class Undefined:
    pass


type TableKey = str | int | Undefined


class TableCell(ABC):
    pass


@dataclass
class StringCell(TableCell):
    s: str
    color: int | None = None
    fixed_width: bool = False
    attr: int = 0


@dataclass
class NumberCell(TableCell):
    v: int | float


def to_table_cell(v: Json | Undefined) -> TableCell:
    match v:
        case False:
            return StringCell("false", curses.COLOR_RED, True)
        case True:
            return StringCell("true", curses.COLOR_GREEN, True)
        case float() | int():
            return NumberCell(v)
        case "":
            return StringCell('""', attr=curses.A_DIM)
        case str():
            encoded = json.dumps(v, ensure_ascii=False)
            if v != encoded[1:-1]:
                v = encoded
            return StringCell(v)
        case []:
            return StringCell(
                "[]",
                curses.COLOR_MAGENTA,
                True,
                curses.A_DIM,
            )
        case list():
            return StringCell("[…]", curses.COLOR_MAGENTA, True)
        case dict() if not v:
            return StringCell(
                "{}",
                curses.COLOR_MAGENTA,
                True,
                curses.A_DIM,
            )
        case dict():
            return StringCell("{…}", curses.COLOR_MAGENTA, True)
        case None:
            return StringCell("null", curses.COLOR_YELLOW, True, curses.A_DIM)
        case Undefined():
            return StringCell("")


def to_dict(v: Json) -> dict[TableKey, TableCell]:
    match v:
        case list():
            return {
                Undefined(): to_table_cell(v),
                **{i: to_table_cell(v) for i, v in enumerate(v)},
            }
        case dict():
            return {
                Undefined(): to_table_cell(v),
                **{k: to_table_cell(v) for k, v in v.items()},
            }
        case _:
            return {Undefined(): to_table_cell(v)}


type TableContent = dict[TableKey, dict[TableKey, TableCell]]


def to_table_content(
    v: Json,
) -> TableContent:
    match v:
        case list():
            return {i: to_dict(v) for i, v in enumerate(v)}
        case dict():
            return {k: to_dict(v) for k, v in v.items()}
        case _:
            return {Undefined(): to_dict(v)}


def compare_prefix_len(base: str | None, a: str, b: str):
    if base is None:
        return True
    for i, c in enumerate(base):
        if i >= len(a) or a[i] != c:
            return False
        if i >= len(b) or b[i] != c:
            return True
    return True


def merge_keys(a_set: set[str], a: list[str], b: list[str]):
    b_set = set(b)
    ia, ib = (0, 0)
    # Using a dict because it preserves insertion order
    res: dict[str, None] = {}
    prev_key = None
    while True:
        if ia >= len(a):
            return list(res.keys()) + b[ib:]
        if ib >= len(b):
            return list(res.keys()) + a[ia:]

        if a[ia] in res:
            ia += 1
        elif b[ib] in res:
            ib += 1
        elif a[ia] == b[ib]:
            res[a[ia]] = None
            prev_key = a[ia]
            ia += 1
            ib += 1
        elif b[ib] in a_set:
            res[a[ia]] = None
            prev_key = a[ia]
            ia += 1
        elif a[ia] in b_set:
            res[b[ib]] = None
            prev_key = b[ib]
            ib += 1
        elif compare_prefix_len(prev_key, a[ia], b[ib]):
            res[a[ia]] = None
            prev_key = a[ia]
            ia += 1
        else:
            res[b[ib]] = None
            prev_key = b[ib]
            ib += 1


def collect_keys(entries: Iterable[Iterable[TableKey]]):
    undefined = set[Undefined]()
    keys_set = set[str]()
    keys_order: list[str] = []
    max_len = 0
    for entry in entries:
        undefined.update({key for key in entry if isinstance(key, Undefined)})
        max_len = max([max_len, *(key + 1 for key in entry if isinstance(key, int))])
        str_keys = [key for key in entry if isinstance(key, str)]
        if any(key not in keys_set for key in str_keys):
            keys_order = merge_keys(keys_set, keys_order, str_keys)
            keys_set.update(str_keys)
    return [*undefined, *keys_order, *range(max_len)]


def integer_digits(v: int | float):
    if int(abs(v)) == 0:
        return 1
    return int(log10(int(abs(v)))) + 1


class JsonCellFormatter(CellFormatter[TableCell]):
    def __init__(self, cells: Iterable[TableCell]):
        # For StringCells
        min_width = 1
        full_width = 1

        # For NumberCells
        negative = False
        integer_width = 1
        fraction_width = None
        exponent_width = 2

        for cell in cells:
            match cell:
                case StringCell():
                    full_width = max(full_width, len(cell.s))
                    if cell.fixed_width:
                        min_width = max(min_width, len(cell.s))
                case NumberCell(v):
                    if v < 0:
                        negative = True
                    exponent_width = max(
                        exponent_width, integer_digits(integer_digits(v))
                    )
                    integer_width = max(integer_width, integer_digits(v))
                    if isinstance(v, float):
                        if fraction_width is None:
                            fraction_width = 1
                        if v != 0.0:
                            fraction_width = max(
                                fraction_width,
                                config.float_precision - integer_digits(v),
                            )
                case _:
                    pass

        self.negative: bool = negative
        self.integer_width: int = integer_width
        self.fraction_width: int | None = fraction_width
        self.exponent_width: int = exponent_width
        self.min_width: int = max(
            min_width,
            negative
            + min(integer_width + 2 * (fraction_width is not None), 5 + exponent_width),
        )
        self.width: int = max(
            full_width, negative + integer_width + 1 + (fraction_width or -1)
        )

    @override
    def draw(
        self,
        cell: TableCell,
        region: Region,
        pos: Point,
        max_width: int | None,
        attr: int,
        force_left: bool,
    ):
        width = self.final_width(max_width)

        def leading_underscores(pos: Point, width: int):
            if not force_left and width > 1:
                region.chgat(
                    pos,
                    width - 1,
                    curses.color_pair(curses.COLOR_BLUE)
                    | curses.A_DIM
                    | curses.A_UNDERLINE
                    | attr,
                )

        match cell:
            case StringCell():
                s = cell.s
                if len(s) > width:
                    s = s[: width - 1] + "…"
                region.insstr(
                    pos,
                    s,
                    cell.attr
                    | (0 if cell.color is None else curses.color_pair(cell.color))
                    | attr,
                )
            case NumberCell(v):
                integer_width = min(
                    self.integer_width,
                    width - self.negative - 2 * (self.fraction_width is not None),
                )
                fraction_width = self.fraction_width and min(
                    self.fraction_width, width - self.negative - integer_width - 1
                )

                if (int_pad := integer_width - integer_digits(v)) >= 0:
                    pad = Point(
                        0, 0 if force_left else (v >= 0 and self.negative) + int_pad
                    )
                    if isinstance(v, int):
                        region.insstr(
                            pos + pad,
                            f"{{:d}}".format(v),
                            curses.color_pair(curses.COLOR_BLUE) | attr,
                        )
                    else:
                        assert fraction_width is not None
                        fraction_width = max(
                            1,
                            min(
                                fraction_width,
                                config.float_precision - integer_digits(v),
                            ),
                        )
                        region.insstr(
                            pos + pad,
                            f"{{:.{fraction_width}f}}".format(v),
                            curses.color_pair(curses.COLOR_BLUE) | attr,
                        )
                    leading_underscores(
                        pos,
                        self.negative + integer_width - (v < 0) - integer_digits(v),
                    )
                else:
                    fmt = f"{{:1.{min(config.float_precision, width-self.negative-4-self.exponent_width)}e}}"
                    try:
                        s = fmt.format(v)
                    except OverflowError:
                        s = fmt.format(Decimal(v))
                    region.insstr(
                        pos + Point(0, (v >= 0 and self.negative) + 0),
                        s,
                        curses.color_pair(curses.COLOR_BLUE) | attr,
                    )

            case _:
                pass


def json_to_table(v: Json) -> Table[int | str | Undefined, TableCell]:
    content = to_table_content(v)

    row_keys = collect_keys([r] for r in content.keys())
    row_headers = [to_table_cell(r) for r in row_keys]
    col_keys = collect_keys(r.keys() for r in content.values())
    col_headers = [to_table_cell(c) for c in col_keys]

    return Table(
        content,
        row_keys,
        row_headers,
        col_keys,
        col_headers,
        [
            JsonCellFormatter(row_headers),
            *(
                JsonCellFormatter(
                    [h, *(r[c] for r in content.values() if c in r)],
                )
                for c, h in zip(col_keys, col_headers)
            ),
        ],
    )



from __future__ import annotations

import curses
import re
from collections.abc import Iterable
from dataclasses import dataclass, field
from typing import Self, override

from tjex.config import config
from tjex.curses_helper import DummyRegion, OffsetRegion, Region
from tjex.history import History
from tjex.kill_ring import KillRing
from tjex.panel import Event, KeyBindings, KeyPress, Panel, StatusUpdate
from tjex.point import Point


class TextPanel(Panel):
    def __init__(self, content: str, clear_first: bool = False):
        self.region: Region = DummyRegion()
        self.content: str = content
        self.attr: int = 0
        self.clear_first: bool = clear_first

    @override
    def resize(self, region: Region):
        self.region = region

    @override
    def handle_key(self, key: KeyPress):
        return [key]

    @override
    def draw(self):
        if self.clear_first:
            for i in range(self.region.height):
                self.region.insstr(Point(i, 0), self.region.width * " ")
        for i, s in enumerate(self.content.splitlines()):
            self.region.insstr(Point(i, 0), s, self.attr)


@dataclass(frozen=True)
class TextEditPanelState:
    content: str
    cursor: int = field(compare=False)


class TextEditPanel(Panel):
    bindings: KeyBindings[Self, None | Event] = KeyBindings()
    word_char_pattern: re.Pattern[str] = re.compile(r"[0-9a-zA-Z_-]")

    def __init__(self, content: str):
        self.region: OffsetRegion = OffsetRegion(DummyRegion(), Point.ZERO)
        self.content: str = content
        self.cursor: int = len(content)
        self.history: History[TextEditPanelState] = History(self.state)
        self.kill_ring: KillRing = KillRing()
        # If the last command was yank or rotate, the position where that yank started, None otherwise
        self.yank_start: int | None = None

    @override
    def resize(self, region: Region):
        self.region = OffsetRegion(region, self.region.offset)

    def next_word(self):
        next_cursor = self.cursor
        while next_cursor < len(self.content) and not self.word_char_pattern.fullmatch(
            self.content[next_cursor]
        ):
            next_cursor += 1
        while next_cursor < len(self.content) and self.word_char_pattern.fullmatch(
            self.content[next_cursor]
        ):
            next_cursor += 1
        return next_cursor

    def prev_word(self):
        next_cursor = self.cursor - 1
        while next_cursor >= 0 and not self.word_char_pattern.fullmatch(
            self.content[next_cursor]
        ):
            next_cursor -= 1
        while next_cursor >= 0 and self.word_char_pattern.fullmatch(
            self.content[next_cursor]
        ):
            next_cursor -= 1
        return next_cursor + 1

    def delete(self, until: int, kill: bool = False):
        until = max(0, min(until, len(self.content)))
        if kill:
            self.kill_ring.kill(
                self.content[min(self.cursor, until) : max(self.cursor, until)],
                prepend=self.cursor > until,
            )
        self.update(
            TextEditPanelState(
                self.content[: min(self.cursor, until)]
                + self.content[max(self.cursor, until) :],
                min(until, self.cursor),
            ),
            killing=kill,
        )

    @bindings.add("C-_")
    def undo(self):
        self.set_state(self.history.pop(self.state))

    @bindings.add("M-_")
    def redo(self):
        self.set_state(self.history.redo())

    @bindings.add("M-w")
    def copy(self):
        """Copy current prompt to clipboard"""
        config.do_copy(self.content)
        return StatusUpdate("Copied.")

    @bindings.add("C-k")
    def kill_line(self):
        """Delete everything to the right of the cursor"""
        self.delete(len(self.content), kill=True)

    @bindings.add("KEY_DC", "C-d")
    def delete_next_char(self):
        self.delete(self.cursor + 1)

    @bindings.add("M-KEY_DC", "M-d", "C-<delete>")
    def delete_next_word(self):
        self.delete(self.next_word(), kill=True)

    @bindings.add("KEY_BACKSPACE")
    def delete_prev_char(self):
        self.delete(self.cursor - 1)

    @bindings.add("M-KEY_BACKSPACE", "C-<backspace>")
    def delete_prev_word(self):
        self.delete(self.prev_word(), kill=True)

    @bindings.add("KEY_RIGHT", "C-f")
    def forward_char(self):
        self.kill_ring.kill_done()
        self.yank_start = None
        self.set_cursor(self.cursor + 1)

    @bindings.add("M-KEY_RIGHT", "C-<right>", "M-<right>", "M-f")
    def forward_word(self):
        self.kill_ring.kill_done()
        self.yank_start = None
        self.set_cursor(self.next_word())

    @bindings.add("KEY_LEFT", "C-b")
    def backward_char(self):
        self.kill_ring.kill_done()
        self.yank_start = None
        self.set_cursor(max(self.cursor - 1, 0))

    @bindings.add("M-KEY_LEFT", "C-<left>", "M-<left>", "M-b")
    def backward_word(self):
        self.kill_ring.kill_done()
        self.yank_start = None
        self.set_cursor(self.prev_word())

    @bindings.add("KEY_END", "C-e")
    def end(self):
        self.kill_ring.kill_done()
        self.yank_start = None
        self.set_cursor(len(self.content))

    @bindings.add("KEY_HOME", "C-a")
    def home(self):
        self.kill_ring.kill_done()
        self.yank_start = None
        self.set_cursor(0)

    @bindings.add("C-y")
    def yank(self):
        pre = self.content[: self.cursor] + self.kill_ring.yank()
        self.yank_start = self.cursor
        self.update(
            TextEditPanelState(
                pre + self.content[self.cursor :],
                len(pre),
            ),
            yanking=True,
        )

    @bindings.add("M-y")
    def rotate(self):
        if self.yank_start is not None:
            pre = self.content[: self.yank_start] + self.kill_ring.rotate()
            self.update(
                TextEditPanelState(
                    pre + self.content[self.cursor :],
                    len(pre),
                ),
                yanking=True,
            )

    @override
    def handle_key(self, key: KeyPress) -> Iterable[Event]:
        match self.bindings.handle_key(key, self):
            case None:
                return ()
            case KeyPress(key_str) if (
                len(key_str) == 1 and key_str not in "\n" and key_str.isprintable()
            ):
                self.kill_ring.kill_done()
                self.yank_start = None
                self.content = (
                    self.content[: self.cursor] + key_str + self.content[self.cursor :]
                )
                self.set_cursor(self.cursor + 1)
            case Event() as event:
                return (event,)
        return ()

    def set_cursor(self, cursor: int):
        self.cursor = max(0, min(len(self.content), cursor))
        self.update_content_base()

    def update_content_base(self):
        if self.cursor < self.region.offset.x:
            self.region.offset = Point(0, self.cursor)
        if self.cursor >= self.region.offset.x + self.region.width:
            self.region.offset = Point(0, self.cursor - self.region.width + 1)
        if len(self.content) < self.region.offset.x + self.region.width:
            self.region.offset = Point(
                0, max(0, len(self.content) - self.region.width + 1)
            )

    @override
    def draw(self):
        self.region.insstr(Point(0, 0), self.content)
        if self.active:
            self.region.chgat(Point(0, self.cursor), 1, curses.A_REVERSE)

    def update(
        self,
        state: str | TextEditPanelState,
        killing: bool = False,
        yanking: bool = False,
    ):
        if isinstance(state, str):
            state = TextEditPanelState(state, len(state))
        self.history.push(self.state)
        self.set_state(state, killing=killing, yanking=yanking)
        self.history.push(self.state)

    @property
    def state(self):
        return TextEditPanelState(self.content, self.cursor)

    def set_state(
        self, state: TextEditPanelState, killing: bool = False, yanking: bool = False
    ):
        if not killing:
            self.kill_ring.kill_done()
        if not yanking:
            self.yank_start = None
        self.content = state.content
        self.set_cursor(state.cursor)



from __future__ import annotations

import importlib.resources
import json
import re
import subprocess as sp
from dataclasses import dataclass
from multiprocessing import Process, Queue, get_start_method
from pathlib import Path
from queue import Empty

from tjex.config import Config, config
from tjex.json_table import Json, TableCell, TableKey, Undefined, json_to_table
from tjex.table import Table
from tjex.utils import TjexError


@dataclass
class JqResult:
    message: str
    table: Table[TableKey, TableCell] | None


class JqError(TjexError):
    pass


identifier_pattern = re.compile(r"[a-zA-Z_][a-zA-Z0-9_]*")


def key_to_selector(key: TableKey):
    match key:
        case Undefined():
            return ""
        case str() if identifier_pattern.fullmatch(key):
            return f".{key}"
        case _:
            return f"[{json.dumps(key, ensure_ascii=False)}]"


def keys_to_selector(*keys: TableKey):
    return "".join(key_to_selector(key) for key in keys)


selector_pattern = re.compile(
    r"""\s*(\.\[("[^\]"\\]*"|\d+)\]|.[a-zA-Z_][a-zA-Z0-9_]*)"""
    + r"""(\.?\[("[^\]"\\]*"|\d+)\]|.[a-zA-Z_][a-zA-Z0-9_]*)*\s*"""
)


def append_filter(command: str, filter: str):
    if command == "":
        return filter
    return command + " | " + filter


def standalone_selector(selector: str):
    return ("" if selector.startswith(".") else ".") + selector


def append_selector(command: str, selector: str):
    if command == "":
        return standalone_selector(selector)
    if selector_pattern.fullmatch(command.split("|")[-1]):
        return command + selector
    return command + " | " + standalone_selector(selector)


class Jq:
    command: str | None = None
    result: Queue[JqResult] | None = None
    process: Process | None = None
    latest_status: JqResult = JqResult("...", None)

    def __init__(self, file: list[Path], slurp: bool):
        # The default start_method "fork" breaks curses
        assert get_start_method() == "forkserver"
        self.file: list[Path] = file
        self.extra_args: list[str] = ["--slurp"] if slurp or len(file) > 1 else []
        self.prelude: str = importlib.resources.read_text(
            "tjex.resources", "builtins.jq"
        )

    @staticmethod
    def run(command: list[str], result: Queue[JqResult], _config: Config):
        # Update global config in subprocess
        for k, v in vars(_config).items():
            setattr(config, k, v)
        try:
            res = sp.run(
                command,
                capture_output=True,
            )
            if res.returncode == 0:
                data: Json = json.loads(res.stdout.decode("utf8"))
                if data is None:
                    result.put(JqResult("null", None))
                else:
                    result.put(JqResult("", json_to_table(data)))
            else:
                result.put(JqResult(res.stderr.decode("utf8"), None))
        except BaseException as e:
            result.put(JqResult(str(e), None))

    def update(self, command: str, force: bool = False):
        if force or command != self.command:
            if self.process is not None:
                self.process.terminate()
                self.process.join()
                self.process.close()
            if self.result is not None:
                self.result.close()
            self.result = Queue()

            self.process = Process(
                target=self.run,
                args=(
                    [
                        config.jq_command,
                        *self.extra_args,
                        self.prelude + (command or "."),
                        *self.file,
                    ],
                    self.result,
                    config,
                ),
            )
            self.process.start()
            self.command = command

    def status(self, block: bool = False, timeout: float = 2) -> JqResult | None:
        if self.result is None:
            return None
        try:
            self.latest_status = self.result.get(block, timeout)
            if self.process is not None:
                self.process.join()
                self.process.close()
                self.process = None
            self.result.close()
            self.result = None
        except Empty:
            self.latest_status = JqResult("...", None)
        return self.latest_status

    def run_plain(self, command: str | None = None) -> Json:
        if command is None:
            command = self.command
        res = sp.run(
            [
                config.jq_command,
                *self.extra_args,
                self.prelude + (command or "."),
                *self.file,
            ],
            capture_output=True,
        )
        if res.returncode != 0:
            raise JqError(res.stderr.decode("utf8"))
        return json.loads(res.stdout.decode("utf8"))



from __future__ import annotations


class KillRing:
    def __init__(self):
        self.ring: list[str] = []
        self.kill_active: bool = False
        self.yank_index: int = 0

    def kill(self, v: str, prepend: bool = False):
        if self.kill_active:
            self.ring[-1] = v + self.ring[-1] if prepend else self.ring[-1] + v
        else:
            self.ring.append(v)
        self.kill_active = True
        self.yank_index = len(self.ring) - 1

    def kill_done(self):
        self.kill_active = False

    def yank(self):
        if self.ring:
            return self.ring[self.yank_index]
        return ""

    def rotate(self):
        self.yank_index = (self.yank_index - 1) % max(1, len(self.ring))
        return self.yank()



# PYTHON_ARGCOMPLETE_OK
from __future__ import annotations

import argparse
import curses
import json
import os
import shlex
import subprocess as sp
import sys
import time
from dataclasses import dataclass
from importlib.metadata import version
from multiprocessing import set_start_method
from pathlib import Path
from typing import Any, Callable

import argcomplete

from tjex import curses_helper, logging
from tjex.config import config as loaded_config
from tjex.config import load as load_config
from tjex.curses_helper import DummyRegion, KeyReader, Region, SubRegion, WindowRegion
from tjex.jq import (
    Jq,
    JqResult,
    append_filter,
    append_selector,
    key_to_selector,
    keys_to_selector,
    standalone_selector,
)
from tjex.json_table import TableCell, TableKey, Undefined
from tjex.logging import logger
from tjex.panel import Event, KeyBindings, KeyPress, StatusUpdate
from tjex.point import Point
from tjex.table_panel import TablePanel, TableState
from tjex.text_panel import TextEditPanel, TextPanel
from tjex.utils import TjexError, TmpFiles


def append_history(jq_cmd: str) -> StatusUpdate:
    skip = False
    cmd: list[str] = ["tjex"]
    for arg in sys.argv[1:]:
        if skip:
            skip = False
        else:
            if arg in {"-c", "--command"}:
                skip = True
            elif arg.startswith("-c") or arg.startswith("--command"):
                pass
            else:
                cmd.append(arg)
    cmd += ["--command", jq_cmd]
    cmd_str = " ".join(shlex.quote(arg) for arg in cmd)

    logger.debug(f"Trying to add to atuin history: {cmd_str}")
    result = sp.run(
        [
            "bash",
            "-c",
            loaded_config.append_history_command.format(shlex.quote(cmd_str)),
        ],
        capture_output=True,
    )
    if result.returncode:
        return StatusUpdate(result.stderr.decode("utf8"))
    return StatusUpdate("Added to atuin history.")


@dataclass
class Quit(Event):
    pass


def tjex(
    screen: Region,
    key_reader: Callable[[], str | None],
    screen_erase: Callable[[], None],
    screen_refresh: Callable[[], None],
    file: list[Path],
    command: str,
    config: Path,
    max_cell_width: int | None,
    slurp: bool,
) -> int:
    table = TablePanel[TableKey, TableCell](screen)
    prompt_head = TextEditPanel("> ")
    prompt = TextEditPanel(command)
    status = TextPanel("")
    status_detail_region = SubRegion(DummyRegion(), Point.ZERO, Point.ZERO)
    status_detail = TextPanel("", clear_first=True)
    status_detail.attr = curses.A_DIM
    panels = [table, prompt_head, prompt, status, status_detail]

    def resize(status_detail_height: None | int = None):
        nonlocal status_detail_region
        status_height = 1
        screen.resize()
        size = screen.size
        table.resize(SubRegion(screen, Point(0, 0), size - Point(3, 0)))
        prompt_head.resize(
            SubRegion(screen, Point(size.y - status_height - 1, 0), Point(1, 2))
        )
        prompt.resize(
            SubRegion(
                screen,
                Point(size.y - status_height - 1, 2),
                Point(1, size.x - 2),
            )
        )
        status.resize(
            SubRegion(
                screen,
                Point(size.y - status_height, 0),
                Point(status_height, size.x),
            )
        )
        if status_detail_height is None:
            status_detail_height = status_detail_region.height
        status_detail_region = SubRegion(
            screen,
            Point(size.y - status_height - 1 - status_detail_height, 0),
            Point(status_detail_height, size.x),
        )
        status_detail.resize(status_detail_region)

    jq = Jq(file, slurp)

    current_command: str = command
    table_cursor_history: dict[str, TableState] = {}

    def set_status(msg: str):
        logger.debug(msg)
        lines = msg.splitlines()
        status.content = "\n".join(lines[:1])
        if len(lines) <= 1:
            resize(status_detail_height=0)
            status_detail.content = ""
        else:
            resize(status_detail_height=len(lines))
            status_detail.content = "\n".join(lines)

    def update_jq_status(block: bool = False):
        nonlocal current_command
        match jq.status(block):
            case JqResult(msg, content):
                if msg == "...":
                    # If result is pending, don't clear previous error message
                    status.content = msg
                else:
                    set_status(msg)
                if content is not None and jq.command is not None:
                    table_cursor_history[current_command] = table.state
                    current_command = jq.command
                    table.update(content, table_cursor_history.get(current_command))
                return True
            case _:
                return False

    bindings: KeyBindings[None, Event | None] = KeyBindings()

    @bindings.add("C-g", "C-d")
    def quit(_: None):  # pyright: ignore[reportUnusedFunction]
        return Quit()

    @bindings.add("M-o", "C-o")
    def toggle_active(_: None):  # pyright: ignore[reportUnusedFunction]
        """Toggle active panel between prompt and table"""
        if active_cycle[0] == prompt:
            update_jq_status(block=True)  # pyright: ignore[reportUnusedCallResult]
        if active_cycle[0] != prompt or jq.latest_status.table is not None:
            active_cycle.append(active_cycle.pop(0))
            active_cycle[-1].set_active(False)
            active_cycle[0].set_active(True)

    _ = bindings.add("C-_", "ESC", name="undo")(lambda _: prompt.undo())
    _ = bindings.add("M-_", name="redo")(lambda _: prompt.redo())

    @bindings.add("M-\n")
    def add_to_history(_: None):  # pyright: ignore[reportUnusedFunction]
        """Append tjex call with current command to shell's history"""
        return append_history(prompt.content)

    @bindings.add("g", "\n")
    def reload(_: None):  # pyright: ignore[reportUnusedFunction]
        """Re-run the current filter."""
        jq.update(prompt.content, force=True)

    @table.bindings.add("M-w")
    def copy_content(_: Any):  # pyright: ignore[reportUnusedFunction]
        """Copy output of current command to clipboard"""
        loaded_config.do_copy(json.dumps(jq.run_plain(), ensure_ascii=False))
        return StatusUpdate("Copied.")

    @table.bindings.add("\n")
    def enter_cell(_: Any):  # pyright: ignore[reportUnusedFunction]
        """Enter highlighted cell by appending selector to jq prompt"""
        prompt.update(
            append_selector(prompt.content, keys_to_selector(*table.cell_keys))
        )

    @table.bindings.add("M-\n")
    def enter_row(_: Any):  # pyright: ignore[reportUnusedFunction]
        """Enter highlighted cell's row by appending selector to jq prompt"""
        prompt.update(append_selector(prompt.content, keys_to_selector(table.row_key)))

    @table.bindings.add("w")
    def copy_cell_content(_: Any):  # pyright: ignore[reportUnusedFunction]
        """Copy content of the current cell to clipboard.
        If content is a string, copy the plain value, not the json representation.
        """
        content = jq.run_plain(
            append_selector(jq.command or ".", keys_to_selector(*table.cell_keys) or "")
        )
        if isinstance(content, str):
            loaded_config.do_copy(content)
        else:
            loaded_config.do_copy(json.dumps(content, ensure_ascii=False))
        return StatusUpdate("Copied.")

    def prompt_append(command: str, cursor: TableState | None = None):
        new_prompt = append_filter(prompt.content, command)
        if cursor is not None:
            table_cursor_history[new_prompt] = cursor
        prompt.update(new_prompt)

    @table.bindings.add("E")
    def expand_row(_: Any):  # pyright: ignore[reportUnusedFunction]
        """Expand the selected row"""
        key = table.row_key
        if key == Undefined():
            raise TjexError("Not an array or object")
        prompt_append(f"expand({json.dumps(key, ensure_ascii=False)})", table.state)

    @table.bindings.add("e")
    def expand_col(_: Any):  # pyright: ignore[reportUnusedFunction]
        """Expand the selected column"""
        key = table.col_key
        if key == Undefined():
            raise TjexError("Not an array or object")
        prompt_append(f"map_values(expand({json.dumps(key, ensure_ascii=False)}))", table.state)

    @table.bindings.add("K")
    def delete_row(_: Any):  # pyright: ignore[reportUnusedFunction]
        """Delete the selected row"""
        key = table.row_key
        if key == Undefined():
            raise TjexError("Not an array or object")
        prompt_append(f"del({standalone_selector(key_to_selector(key))})", table.state)

    @table.bindings.add("k")
    def delete_col(_: Any):  # pyright: ignore[reportUnusedFunction]
        """Delete the selected column"""
        key = table.col_key
        if key == Undefined():
            raise TjexError("Not an array or object")
        prompt_append(
            f"map_values(del({standalone_selector(key_to_selector(key))}))", table.state
        )

    @table.bindings.add("m")
    def select_col(_: Any):  # pyright: ignore[reportUnusedFunction]
        """Enter the selected column"""
        key = table.col_key
        if key == Undefined():
            raise TjexError("Not an array or object")
        prompt_append(
            f"map_values({standalone_selector(key_to_selector(key))})",
            table.state.row_only,
        )

    @table.bindings.add("s")
    def sort_by_col(_: Any):  # pyright: ignore[reportUnusedFunction]
        """Sort rows by the selected column.
        Works only for arrays right now.
        """
        if not isinstance(table.row_key, int):
            raise TjexError("Not an array")
        key = table.col_key
        if key == Undefined():
            prompt.update(append_filter(prompt.content, f"sort"))
        else:
            prompt_append(
                f"sort_by({standalone_selector(key_to_selector(key))})",
                table.state.col_only,
            )

    load_config(
        config, {"global": bindings, "prompt": prompt.bindings, "table": table.bindings}
    )
    if max_cell_width:
        loaded_config.max_cell_width = max_cell_width

    resize()
    active_cycle = [table, prompt]
    if loaded_config.start_at_prompt:
        active_cycle = [prompt, table]
    active_cycle[0].set_active(True)
    jq.update(prompt.content)

    redraw = True

    while True:
        if (key := key_reader()) is not None:
            try:
                for event in active_cycle[0].handle_key(KeyPress(key)):
                    match bindings.handle_key(event, None):
                        case Quit():
                            return 0
                        case StatusUpdate(msg):
                            set_status(msg)
                        case KeyPress("KEY_RESIZE"):
                            resize()
                        case _:
                            pass
            except TjexError as e:
                set_status(e.msg)
            jq.update(prompt.content)
            redraw = True
            continue

        if update_jq_status() or redraw:
            screen_erase()
            for panel in panels:
                panel.draw()
            screen_refresh()
            redraw = False
            continue

        time.sleep(0.01)


def main():
    set_start_method("forkserver")
    parser = argparse.ArgumentParser(description="A tabular json explorer.")
    _ = parser.add_argument("--version", action="version", version=version("tjex"))
    _ = parser.add_argument("file", type=Path, nargs="*")
    _ = parser.add_argument("-c", "--command", default="")
    _ = parser.add_argument(
        "--config", type=Path, default=Path.home() / ".config" / "tjex" / "config.toml"
    )
    _ = parser.add_argument("--logfile", type=Path)
    _ = parser.add_argument("-w", "--max-cell-width", type=int)
    _ = parser.add_argument("-s", "--slurp", action="store_true")
    argcomplete.autocomplete(parser)
    args = parser.parse_args()
    logging.setup(args.logfile)

    with TmpFiles() as tmpfile:
        if not args.file:
            args.file = [tmpfile(sys.stdin.read())]
            os.close(0)
            sys.stdin = open("/dev/tty")
        for i in range(len(args.file)):
            if not args.file[i].is_file():
                args.file[i] = tmpfile(args.file[i].read_text())

        @curses.wrapper
        def result(scr: curses.window):
            _ = curses.curs_set(0)
            curses_helper.setup_plain_colors()
            return tjex(
                WindowRegion(scr),
                KeyReader(scr).get,
                scr.erase,
                scr.refresh,
                **{n: k for n, k in vars(args).items() if n not in {"logfile"}},
            )

    return result


if __name__ == "__main__":
    exit(main())



from contextlib import ExitStack, contextmanager
from pathlib import Path
from tempfile import NamedTemporaryFile


class TjexError(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)
        self.msg: str = msg


@contextmanager
def TmpFiles():
    with ExitStack() as stack:

        def tmpfile(s: str):
            buffered = stack.enter_context(
                NamedTemporaryFile(mode="w", delete_on_close=False, delete=True)
            )
            _ = buffered.write(s)
            buffered.close()
            return Path(buffered.name)

        yield tmpfile
